;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : environment-cmds.lisp
;;; Version     : 1.5
;;; 
;;; Description : No system dependent code.
;;;             : Defines stuff that is needed to handle "environment"
;;;             : things.
;;; Bugs        : 
;;; 
;;; Todo        : Clean up what goes here and what's passed literally
;;;             : from the Tcl side - it's an odd mix at this point
;;;             : and I should have some consistency.
;;; 
;;; ----- History -----
;;;
;;; 05/22/2002  Dan
;;;             : File creation
;;; 10/01/2002  Dan
;;;             : Updated version to 1.1 and fixed the packaging
;;;             : for building a standalone in ACL.
;;; 01/20/2003  Dan
;;;             : Added the show-module-state-chunk function and
;;;             : hacked buffer-contents so that the module-state
;;;             : chunks shown by the environment match internal
;;;             : module states even though the "real" chunks
;;;             : don't seem to be updated anymore. WHY and WHEN?!
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

;;; Here I define a variable that will be used by the environment
;;; as a simple flag to prevent run from being pressed more than
;;; once (and there may be other uses)

(defvar *running-actr* nil)

;;; Reset-model
;;; this is the function called when the reset button in the
;;; environment is pressed.  Basically, all it does print a message
;;; after it resets for reference.

(defun reset-model ()
  (reset)
  (format t "~%#|## Model has been reset. ##|#~%"))

;;; Reload-model
;;; this is the function called when the reload button in the
;;; environment is pressed.  It takes one parameter which specifies 
;;; whether or not to use the smart-loader instead of the simple reload
;;; function.

(defun reload-model (smart-load?)
  (if smart-load?
      (let ((result (smart-loader (first *model*))))
        (when (= 1 (first result))
          (format t "~%#|## Model has been reloaded. ##|#~%"))
        result)
    (let* ((save-stream (make-string-output-stream ))
           (*standard-output* save-stream)
           (*error-output* save-stream))
      (multiple-value-bind (s err) 
          (ignore-errors (reload))
        (declare (ignore s))
        (if (and (subtypep (type-of err) 'condition)
                 (not (equal (type-of err) 'unbound-variable)))
            (progn
              (uni-report-error err 
                                (format nil "Error reloading model"))
              (list 0 (get-output-stream-string save-stream)))
          (progn
            (format t "~%#|## Model has been reloaded. ##|#~%")
            (list 1 (get-output-stream-string save-stream))))))))


;;; buffer-list
;;; This function doesn't take any parameters, and 
;;; doesn't really do anything at this point other than
;;; return a constant list. 
;;; I'm hoping that with 6.0 there'll be a way to add and
;;; access buffers uniformly (unlike now) and there'll exist
;;; a function for getting the list, but for now just hack
;;; it out.

(defun buffer-list (x)
  (declare (ignore x))
  '(goal retrieval visual-location visual visual-state
         aural-location aural aural-state
         manual-state vocal-state) ; manual vocal ???
  )

;;; buffer-contents
;;; This function takes one parameter which is a symbol that
;;; is the name of a buffer.  It prints out the chunk that is
;;; currently in that buffer, or "Buffer is Empty" if there isn't
;;; a chunk in that buffer to *standard-output*.
;;; This is another of those mechanisms that'd be nice to fix
;;; for ACT-R 6.0...
;;; (Does anybody bother to read any of this stuff?)

(defun buffer-contents (buffer)
  (case buffer
    (goal (if (car (buffers goal))
              (goal-focus)
            (format t "Buffer is Empty")))
    (retrieval (if (car (buffers retrieval))
                   (wm-fct (buffers retrieval))
                 (format t "Buffer is Empty")))
    
    (visual-location (if (car (buffers visual-location))
                         (wm-fct (buffers visual-location))
                       (format t "Buffer is Empty")))
    (aural-location (if (car (buffers aural-location))
                         (wm-fct (buffers aural-location))
                       (format t "Buffer is Empty")))
    (visual-state (show-module-state-chunk (vis-m *mp*)))
    (manual-state (show-module-state-chunk (motor-m *mp*)))
    (vocal-state (show-module-state-chunk (speech-m *mp*)))
    (aural-state (show-module-state-chunk (audio-m *mp*)))
    (visual (if (currently-attended (vis-m *mp*))
                (wm-fct (list (id (currently-attended (vis-m *mp*)))))
              (format t "Buffer is Empty")))
    (aural (if (currently-attended (audio-m *mp*))
                (wm-fct (list (id (currently-attended (audio-m *mp*)))))
              (format t "Buffer is Empty")))
    
    (t nil)))

;;; show-module-state-chunk
;;; This function takes 1 parameter which is the module for which
;;; state is desired.
;;; It's a hack for now because John noticed the problem and I can't
;;; figure out where or when the module-state chunks stopped 
;;; properly reflecting internal module states.

(defun show-module-state-chunk (mod)
  (wm-fct (list (ps-ptr (state-dmo mod)))))


;;; safe-whynot
;;; This function takes any number of production names as parameters.
;;; It calls the ACT-R whynot function after protecting the globals that
;;; it 'smashes' because that really causes problems precisely when it's
;;; most useful to use whynot - when the instantiation-hook is looping in
;;; the stepper so that the system is suspended.  
;;; For each production passed to it the whynot function prints out either 
;;; why the production doesn't match current state, or the full instantiation
;;; if it does match.

(defun safe-whynot (&rest prods)                       
  (let ((*conflict-set* nil)
        (*instantiation* nil)
        (*extra-instantiation* nil))
    (whynot-fct prods)))

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#