;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : hook-fns.lisp
;;; Version     : 1.5
;;; 
;;; Description : NO system dependent code.
;;;             : This file contains the functions and variables 
;;;             : for splitting a model file into pieces which are
;;;             : saved in the environment directory.
;;; Bugs        : 
;;; 
;;; Todo        :
;;; 
;;; ----- History -----
;;;
;;; 10/01/2002  Dan
;;;             : Added the header.
;;; 12/05/2002  Dan
;;;             : Hacked in the create-valid-pathname call to
;;;             : play friendly in MCL 5 because the Tcl side
;;;             : uses Unix pathnames but MCL wants Mac pathnames.
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

;;; using fixed file names - note this means that there can effectively be only
;;; one environment running on a machine.

(defvar *chunk-type-file* (merge-pathnames "chunk-type.lisp" *load-pathname*))
(defvar *chunk-file* (merge-pathnames "chunk.lisp" *load-pathname*))
(defvar *production-file* (merge-pathnames "production.lisp" *load-pathname*))
(defvar *misc-file* (merge-pathnames "misc.lisp" *load-pathname*))
(defvar *command-file* (merge-pathnames "command.lisp" *load-pathname*))


;;; split-model-file
;;; This function takes one parameter which should be a file name.
;;; It opens that file and parses it into the environment model part files.
;;; It doesn't check for syntax errors in the file because it is assumed 
;;; to be a "well formed" Lisp file because the only time it gets called
;;; is after a successful load of that file.

(defun split-model-file (fname)
  (setf fname (create-valid-pathname fname))
   (let ((chunk-type nil)
        (chunk nil)
        (production nil)
        (misc nil)
        (command nil)
        (top-params nil)
        (bottom-params nil))
    
    (with-open-file (in-file fname :direction :input)
      (do* ((file-len (file-length in-file))
            (start 0 end)
            (form (read in-file nil 'error-reading-actr-environment-file)
                  (read in-file nil 'error-reading-actr-environment-file))
            
            (cmd (if (listp form) (car form) nil) 
                 (if (listp form) (car form) nil))
            
            (end (if (equal form 'error-reading-actr-environment-file)
                     file-len
                   (file-position in-file)) 
                 (if (equal form 'error-reading-actr-environment-file)
                     file-len
                   (file-position in-file))))
           
           ((equal form 'error-reading-actr-environment-file)  
            (unless (= start file-len)
              (push (list start end) bottom-params)))
        
        (cond ((eql cmd 'p) 
               (push (list start end) production))
              
              ((or (eql cmd 'wmetype)
                   (eql cmd 'chunk-type))      
               (push (list start end) chunk-type))
              
              ((or (eql cmd 'set-dm)
                   (eql cmd 'setwm)
                   (eql cmd 'addwm)
                   (eql cmd 'add-dm))
               (push (list start end) chunk))
              
              ((or (eql cmd 'clearall)
                   (eql cmd 'clear-all)
                   (eql cmd 'pm-reset))
               ;; just eat clearall and pm-reset, but should do something
               ;; about warning whether they exist or not but unsure how/where
               ;; right now
               )
              ((or (eql cmd 'sgp) 
                   (eql cmd 'sgp-fct)
                   (eql cmd 'pm-set-params)
                   (eql cmd 'define-buffer))
               (push (list start end) top-params))
              
              ((or (eql cmd 'spp)
                   (eql cmd 'spp-fct)
                   (eql cmd 'parameters))
               (push (list start end) bottom-params))
              
              ((or (eql cmd 'load)
                   (eql cmd 'load-model)
                   )
               (push (list start end) top-params)
               ;; old environment ate these and put the files into the
               ;; model manager, but I'm unsure if I want to do that now
               ;; so for now I'll put them at the "top"
               )
              
              ((eql cmd 'add-graph-to-model)
               (push (list start end) command)
               ;; don't have a graph manager yet, so just tuck these into the
               ;; commands window for now
               )
              ((member cmd '(defun defmacro defmethod defvar defclass 
                              defparameter defconstant))
               (push (list start end) misc))
              
              (t
               (push (list start end) command)))))
    
    ;; because I'm paranoid about the :overwrite option (I've seen it
    ;; 'not' work right sometimes) I'll delete all the model files first...
    
    (dolist (x (list *chunk-type-file* *chunk-file* *production-file*
                     *misc-file* *command-file*))
      (when (probe-file x)
        (delete-file x)))
    
    (write-file-part fname *chunk-type-file* 
                     (reverse chunk-type) :overwrite)
    (write-file-part fname *chunk-file* (reverse chunk) :overwrite)
    (write-file-part fname *production-file* 
                     (reverse production) :overwrite)
    (write-file-part fname *misc-file* (reverse misc) :overwrite)
    (write-file-part fname *command-file* (reverse top-params) :overwrite)
    (write-file-part fname *command-file* (reverse command) :append)
    (write-file-part fname *command-file* (reverse bottom-params) :append)
  
    ))

(defun write-file-part (in-file-name out-file-name markers method)
  (with-open-file (in-file in-file-name :direction :input)
    (with-open-file (out-file out-file-name :direction :output
                              :if-exists method :if-does-not-exist :create)
      (dolist (x markers)
        (file-position in-file (car x))
        (do ((pos (file-position in-file)
                  (file-position in-file)))
            ((equal pos (second x)))
          (write-char (read-char in-file) out-file)))))) 
  

(defun write-model-from-parts (fname)
  (setf fname (create-valid-pathname fname))
  (let ((problem-list nil))
    (dolist (x (list (cons *chunk-type-file* "chunk-type")
                     (cons *chunk-file* "chunk")
                     (cons *production-file* "production")
                     (cons *misc-file* "misc")
                     (cons *command-file* "command")))
      (multiple-value-bind (s err) 
          (ignore-errors 
           (with-open-file (in-file (car x) :direction :input)
             (do ((form 
                   (read in-file nil 'error-reading-actr-environment-file)
                   (read in-file nil 'error-reading-actr-environment-file)))
                  
                 ((equal form 'error-reading-actr-environment-file)))))
        (declare (ignore s))
        (when (and (subtypep (type-of err) 'condition)
                   (not (equal (type-of err) 'unbound-variable)))
          (push (cdr x) problem-list))))
          
  (when problem-list
    (return-from write-model-from-parts (list 0 (reverse problem-list)))))
  
  (split-model-file fname)
  
  (when (probe-file fname)
    (delete-file fname))
  
  (let ((len (with-open-file (f *misc-file* :direction :input)
               (file-length f))))
    (write-file-part *misc-file* fname (list (list 0 len)) :error))
  
  (with-open-file (f fname :direction :output :if-exists :append)
    (format f "~%(clear-all)~%(pm-reset)~%"))
  
  (let ((command nil)
        (top-params nil)
        (bottom-params nil))
    
    (with-open-file (in-file *command-file* :direction :input)
      (do* ((file-len (file-length in-file))
            (start 0 end)
            (form (read in-file nil 'error-reading-actr-environment-file)
                  (read in-file nil 'error-reading-actr-environment-file))
            
            (cmd (if (listp form) (car form) nil) 
                 (if (listp form) (car form) nil))
            
            (end (if (equal form 'error-reading-actr-environment-file)
                     file-len
                   (file-position in-file)) 
                 (if (equal form 'error-reading-actr-environment-file)
                     file-len
                   (file-position in-file))))
           
           ((equal form 'error-reading-actr-environment-file)  
            (unless (= start file-len)
              (push (list start end) 
                    bottom-params)))
        
        (cond ((or (eql cmd 'sgp) 
                   (eql cmd 'sgp-fct)
                   (eql cmd 'pm-set-params)
                   (eql cmd 'define-buffer))
               (push (list start end) top-params))
              
              ((or (eql cmd 'spp)
                   (eql cmd 'spp-fct)
                   (eql cmd 'parameters))
               (push (list start end) bottom-params))
              
              (t
               (push (list start end) command)))))
    
    (write-file-part *command-file* fname (reverse top-params) :append)
    
    (let ((len (with-open-file (f *chunk-type-file* :direction :input)
                 (file-length f))))
      (write-file-part *chunk-type-file* fname (list (list 0 len)) :append))
    
    (let ((len (with-open-file (f *chunk-file* :direction :input)
                 (file-length f))))
      (write-file-part *chunk-file* fname (list (list 0 len)) :append))
    
    (let ((len (with-open-file (f *production-file* :direction :input)
                 (file-length f))))
      (write-file-part *production-file* fname (list (list 0 len)) :append))

    (write-file-part *command-file* fname (reverse command) :append)
    (write-file-part *command-file* fname (reverse bottom-params) :append))
  (list 1 t))
 
#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#