;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : smart-loader.lisp
;;; Version     : 1.5
;;; 
;;; Description : Some system dependent code (due to odd bugs).
;;;             : Defines the functions that handle loading the models when
;;;             : requested by the Tcl side.  It writes all the defun, 
;;;             : defmethod, defclass, and defstructs from the file to a
;;;             : temp file and then compiles that and loads it.
;;; Bugs        : LispWorks doesn't seem to use compile-file correctly in
;;;             : smart-loader, and OpenMCL also has a problem with it though
;;;             : I'm not sure what that is, so for now the solution is to
;;;             : just not allow it to be used in those systems.
;;; 
;;; Todo        : 
;;; 
;;; ----- History -----
;;;
;;; 06/09/2002  Dan
;;;             : Creation.
;;; 08/20/2002  Dan
;;;             : Added the revised smart-loader function that just issues
;;;             : a warning in LispWorks and OpenMCL.
;;; 09/17/02    Dan
;;;             : Changed the compiled-file's name to be sensitive to the
;;;             : system to fix the problem with OpenMCL and LispWorks.
;;; 10/01/2002  Dan
;;;             : Updated version to 1.1 and fixed the packaging
;;;             : for building a standalone in ACL.
;;; 12/05/2002  Dan
;;;             : Hacked in the create-valid-pathname call to
;;;             : play friendly in MCL 5 because the Tcl side
;;;             : uses Unix pathnames but MCL wants Mac pathnames.
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

(defconstant *forms-to-compile* '(defun defmethod defclass defstruct))

(defvar *temp-source-file* (merge-pathnames "env-temp.lisp" *load-pathname*))
(defvar *temp-compiled-file* (merge-pathnames (merge-pathnames "env-temp" *load-pathname*) *.fasl-pathname*))

(defun safe-load (file)
  (setf file (create-valid-pathname file))
  (let* ((save-stream (make-string-output-stream ))
         (*standard-output* save-stream)
         (*error-output* save-stream))
    (multiple-value-bind (s err) 
        (ignore-errors (load file))
      (declare (ignore s))
      (if (and (subtypep (type-of err) 'condition)
               (not (equal (type-of err) 'unbound-variable)))
          (progn
            (uni-report-error err (format nil "Error loading file: ~S~%" file))
            (list 0 (get-output-stream-string save-stream)))
        (list 1 (get-output-stream-string save-stream))))))
        
#+:ACTR-ENV-ALONE
(defun smart-loader (file)
  (declare (ignore file))
  (list 0 "You must disable the 'Compile definitions when model opened or reloaded' option to be able to open a model"))


#-:ACTR-ENV-ALONE
(defun smart-loader (file)
  (setf file (create-valid-pathname file))
  (let* ((save-stream (make-string-output-stream ))
         (*standard-output* save-stream)
         (*error-output* save-stream))
     (multiple-value-bind (s err) 
        (ignore-errors (load file))
       (declare (ignore s))
       (if (and (subtypep (type-of err) 'condition)
               (not (equal (type-of err) 'unbound-variable)))
          (progn
            (uni-report-error err (format nil "Error loading file: ~S~%" file))
            (list 0 (get-output-stream-string save-stream)))
        (let* ((null-stream (make-string-output-stream ))
               (*standard-output* null-stream)
               (*error-output* null-stream))
          
          ;; for saftey reasons, just delete them first
          (when (probe-file *temp-source-file*)
            (delete-file *temp-source-file*))
          
          (when (probe-file *temp-compiled-file*)
            (delete-file *temp-compiled-file*))
          
          (with-open-file (f file :direction :input)
            (with-open-file (f2 *temp-source-file* :direction :output
                                :if-exists :overwrite
                                :if-does-not-exist :create)
              (do ((form (read f nil 'finished-reading-file)
                       (read f nil 'finished-reading-file)))
                ((equal form 'finished-reading-file))
              (when (and (listp form) (member (car form) *forms-to-compile*))
                (pprint form f2)))))
          (compile-file *temp-source-file* :output-file *temp-compiled-file*)
          (load *temp-compiled-file*)
          (list 1 (get-output-stream-string save-stream)))))))
    
#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#