;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : stepper-control.lisp
;;; Version     : 1.5
;;; 
;;; Description : No system dependent code.
;;;             : This file contains the Lisp to support the stepper window.
;;;             : 
;;; Bugs        : 
;;; 
;;; Todo        : Tutor mode...
;;; 
;;; ----- History -----
;;;
;;; 05/24/2002  Dan
;;;             : File creation
;;; 10/01/2002  Dan
;;;             : Updated version to 1.1 and fixed the packaging
;;;             : for building a standalone in ACL.
;;;             : Modified the stepper-control-function to use the
;;;             : new uni-wait-for function.
;;; 10/14/2002  Dan
;;;             : Made the changes to implement the tutor mode
;;;             : in the instantiation window instead of the bindings
;;;             : window of the stepper.
;;; 11/11/2002  Dan
;;;             : Modified stepper-instan-info and stepper-control-function
;;;             : so that instantiation picking was possible.  It works
;;;             : off of the step buttons value and the *last-stepper-instantiation*
;;;             : variable.  Requires that the option be set in the
;;;             : environment before it's enabled.
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;             : Updated run-master-process with the RPM 2.2 version.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

(defvar *which-stepper-list* nil)

;;; *stepper-events*
;;; This global variable is used to hold the conditions on which
;;; the stepper should report an event.
;;; I had a 'cleaner' mechanism that just pushed the
;;; corresponding keywords onto a list, but I realized that
;;; there was the potential for problems with multiple threads
;;; accessing the same global and I didn't want to get into
;;; semaphores cross paltform - so I've switched it to a 
;;; table, which I don't think suffers from that same fate
;;; since the keys are unique in the separate threads
;;; keys are, :production, :retrieval, :motor, :vision, :audio, :speech,   
;;; :device, or :external

(defvar *stepper-events* (make-hash-table))

;;; *stepper-update-table*
;;; This variable holds the handlers that need to be updated when the
;;; stepper steps.
;;; I'm continuing with the same idea that a hash table won't
;;; suffer the consequences of multiprocess race conditions...

(defvar *stepper-update-table* (make-hash-table))

;;; *stepper-advance*
;;; This global is used to indicate when a step should continue.  It's
;;; set when the step and stop buttons are pressed or if the stepper is closed.

(defvar *stepper-advance* nil)

;;; *stopped-by-stepper*
;;; This global is used to indicate that the stop button was pressed and
;;; that the model should "stop".

(defvar *stopped-by-stepper* nil)

;;; *last-stepper-instantiation*
;;; This global is used to hold the production that was last selected from
;;; the list of productions to support instantiation picking.

(defvar *last-stepper-instantiation* nil)

;;; *true-stepper-instantiation*
;;; This global is used to hold the production that was "selected" 
;;; during the instantiation hook.

(defvar *true-stepper-instantiation* nil)


;;; *stepper-skip-until-time*
;;; This variable is set to either a time or nil and used to indicate whether
;;; or not the stepper should wait for a user requested time before 
;;; stepping any more - essentially it allows for a "fast-forward" of
;;; sorts (given that nothing's fast when the environment is connected).

(defvar *stepper-skip-until-time* nil)

(defvar *tutor-mode* nil)
(defvar *tutor-bindings* nil)
(defvar *tutor-responses* nil)

(defun stepper-continuing ())

(defun schedule-stepper-skip-time (&optional (s-time nil))
  (when s-time
    (if (equal (car s-time) 'time)
        (let ((num (cadr s-time)))
          (when (and (numberp num) (> num (mp-time *mp*)))
            (setf *stepper-skip-until-time* num)
            (setf (sched-q *mp*)
                  (queue-insert (make-instance 'input-queue-entry
                                  :priority 4
                                  :time-tag num
                                  :destination :EXTERNAL
                                  :params (list #'stepper-continuing))
                                (sched-q *mp*)))))
        (let ((prod (cadr s-time)))
          (when (member prod (no-output (pp)))
            (setf *stepper-skip-until-time* prod))
          ))))
    

;;; stepper-control-function
;;; This function is called to update the stepper main handler.  The
;;; parameter will be either the handler, an instantiation, a scheduler queue
;;; entry or nil (not sure why it's ever nil, but sometimes it is).

(defun stepper-control-function (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on the first update clear the stop flag and the table of handlers
      (progn
        (setf *stopped-by-stepper* nil)
        (maphash #'(lambda (key val)  
                     (declare (ignore val))
                     (remhash key *stepper-update-table*)) 
                 *stepper-update-table*))
        
    ;; otherwise if it's an event that needs to be stepped at then
    ;; update all of the handlers and wait for the request to continue.
    
    (when (and (currently-handled-event event))
      (if (eq :post *which-hook-fn*)
          (progn
            (maphash #'(lambda (key val) 
                         (declare (ignore key))
                         (update-handler val nil))
                     *stepper-update-table*)
            (when *stopped-by-stepper*
              ;; need to force it to stop after this event
              ;; so put an event at the head of the queue that
              ;; had better be past the run time...
              (setf (sched-q *mp*)
                (cons (make-instance 'queue-entry 
                      :time-tag 999999999
                      :destination :external)
                    (sched-q *mp*)))))
        
        (progn
          (maphash #'(lambda (key val) 
                       (declare (ignore key))
                       (update-handler val event))
                   *stepper-update-table*)
          
          (uni-wait-for #'(lambda () (or (eq *stepper-advance* :step)
                                         (eq *stepper-advance* :pick)
                                         (eq *stepper-advance* :stop))))
          (when *tutor-bindings*
            (setf *tutor-bindings* nil)
            (setf *tutor-responses* nil))
          (when (eq *stepper-advance* :stop)
            (setf *stopped-by-stepper* t))
          
          (if (and (eq *stepper-advance* :pick) (eq :instan *which-hook-fn*))
              (let ((instan (find (production-name *last-stepper-instantiation*)
                                  *conflict-set* 
                                  :key #'(lambda (x) 
                                           (production-name 
                                            (instantiation-production x))))))
                ; (format t "~%--------~%The instan for ~S is ~S~%---------~%" *last-stepper-instantiation* instan)
                instan)
            (progn
              (setf *last-stepper-instantiation* *true-stepper-instantiation*)
              nil)))))))

;;; stepper-cur-event
;;; This is the function that handles the updates of the next event
;;; window in the stepper.  The event that's passed in will be either
;;; the handler (the creation update), a queue-entry, an instantiation or nil.
;;; It's a text-output handler so the display just needs to be printed to
;;; *standard-output* for it to be displayed.  
;;;
;;; With the new step hook it makes things easier because I actually get
;;; the real event - I don't have to go get it from the module's queue.


(defun stepper-cur-event (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (setf (gethash event *stepper-update-table*) event) 
    
    ;; otherwise it needs to print the output
    
    (if event
        (if (subtypep (type-of event) 'queue-entry)
            (let ((dest (destination event))
                   )
              (format t "Module: ~s~%Time: ~s~%Command:" dest (mp-time *mp*))
              (unless (equal (type-of event) 'queue-entry)
                (format t "~{~%~S~}" (params event))))
          
          ;; it's an instantiation which will be handled by the 
          ;; other displays, so just print "what" it is.
          (format t "Module: ~S~%Time: ~s~%Command:~%~S"
            :cognition (mp-time *mp*) 'select-production))
      ;; If it's nil that means clear the window
      nil)))

;;; stepper-instan-list
;;; This is the function that handles the updates of the instantiation
;;; list in the stepper.  The event that's passed in will be either
;;; the handler (the creation update), a queue-entry, an instantiation list
;;; or nil.

(defun stepper-instan-list (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (progn
        (setf (gethash event *stepper-update-table*) event) 
        nil)
    
    ;; otherwise it needs to return the list 
    
    (if event 
        (if (eq :instan *which-hook-fn*)
            (progn
              (setf *which-stepper-list* 'prod)
              (setf *true-stepper-instantiation* 
                (instantiation-production (car event)))
              (setf *last-stepper-instantiation* 
                (instantiation-production (car event)))
              (mapcar (lambda (x) (instantiation-production x)) event))
          (if (and (eq :step *which-hook-fn*)
                   (eq :cognition (destination event)))
              (let* ((cmd (car (params event))))
                
                (if (equal cmd 'execute-rhs)
                    (progn
                      (setf *which-stepper-list* 'prod)
                      
                      (if *last-stepper-instantiation*
                          (list *last-stepper-instantiation*)
                        (list (instantiation-production (car *conflict-set*)))))
                  (if (equal cmd 'complete-retrieval)
                      (progn
                        (setf *which-stepper-list* 'chunk)
                        
                        (cons (cdr *retrieval-scheduler*)
                              (remove (cdr *retrieval-scheduler*)
                                      (mapcar #'car 
                                        (sort (copy-tree 
                                               (cdr *retrieval-record*))
                                              #'(lambda (x y)
                                                  (if (and (numberp x)
                                                           (numberp y))
                                                      (> x y)
                                                    (if (numberp x)
                                                        t
                                                      nil)))
                                              :key #'cadr)))))
                        nil)))
            nil))
      nil)))





(defun stepper-list-name (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (progn
        (setf (gethash event *stepper-update-table*) event) 
        " ")
    
    ;; otherwise it needs to return the list 
    
    (if event 
        (if (eq :instan *which-hook-fn*)
            "Possible Productions"
          (if (and (eq :step *which-hook-fn*)
                   (eq :cognition (destination event)))
              (let* ((cmd (car (params event))))
                
                (if (equal cmd 'execute-rhs)
                    "Possible Productions"
                  (if (equal cmd 'complete-retrieval)
                      "Possible Chunks"
                        " ")))
            " "))
      " ")))



(defun stepper-details-name (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (progn
        (setf (gethash event *stepper-update-table*) event) 
        " ")
    
    ;; otherwise it needs to return the list 
    
    (if event 
        (if (eq :instan *which-hook-fn*)
            "Production Details"
          (if (and (eq :step *which-hook-fn*)
                   (eq :cognition (destination event)))
              (let* ((cmd (car (params event))))
                
                (if (equal cmd 'execute-rhs)
                    "Production Details"
                  (if (equal cmd 'complete-retrieval)
                      "Chunk Details"
                        " ")))
            " "))
      " ")))



(defun stepper-bindings-name (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (progn
        (setf (gethash event *stepper-update-table*) event) 
        " ")
    
    ;; otherwise it needs to return the list 
    
    (if event 
        (if (eq :instan *which-hook-fn*)
            "Bindings"
          (if (and (eq :step *which-hook-fn*)
                   (eq :cognition (destination event)))
              (let* ((cmd (car (params event))))
                
                (if (equal cmd 'execute-rhs)
                    "Bindings"
                  (if (equal cmd 'complete-retrieval)
                      "Retrieval Request"
                        " ")))
            " "))
      " ")))





(defun stepper-prod_frame-name (event)
  (if (subtypep (type-of event) 'environment-handler)
      ;; on creation need to save it in the table for later use
      (progn
        (setf (gethash event *stepper-update-table*) event) 
        " ")
    
    ;; otherwise it needs to return the list 
    
    (if event 
        (if (eq :instan *which-hook-fn*)
            "Instantiated Production"
          (if (and (eq :step *which-hook-fn*)
                   (eq :cognition (destination event)))
              (let* ((cmd (car (params event))))
                
                (if (equal cmd 'execute-rhs)
                    "Instantiated Production"
                  (if (equal cmd 'complete-retrieval)
                      "Chunk"
                        " ")))
            " "))
      " ")))

;;; stepper-instan-info
;;; This is the function that handles the updates of the instantiation
;;; information box in the stepper.  The parameter passed in is the name of
;;; a production that's chosen in the list (that's the only way this is called)
;;; It prints whether or not this is the "selected" production and it's 
;;; utility.  I may want to display more there later, but for now this seems
;;; like enough.

;;; Hacking this for production "picking" purposes.  Since this gets called when
;;; a production is picked from the list it's as good a place as any to record that
;;; name for later use...


(defun stepper-instan-info (prod)
  (when prod
    (if (equal *which-stepper-list* 'prod)
        (let ((instan (find prod *conflict-set* 
                            :key #'(lambda (x) 
                                     (production-name 
                                      (instantiation-production x))))))
          (setf *last-stepper-instantiation* (instantiation-production instan))
          (if (equal instan (car *conflict-set*))
              (format t "Selected Production~%")
            (format t "NOT Selected Production~%"))
          
          (format t "Utility: ~S" (instantiation-gain instan)))
      (let ((chunk (car (no-output (wm-fct (list prod))))))
        (if (equal chunk (wme-name (cdr *retrieval-scheduler*)))
            (format t "Retrieved Chunk~%")
          (format t "NOT Retrieved Chunk~%"))
        (format t "Matching Score: ~S" 
          (if (equal prod 'failure)
              *retrieval-threshold*
            (cadr (assoc prod (cdr *retrieval-record*) :key #'wme-name))))))))
  
;;; stepper-instan-binding
;;; This is the function that handles the updates of the instantiation
;;; bindings box in the stepper.  The parameter passed in is the name of
;;; a production that's chosen in the list (that's the only way this is called)
;;; It prints the bindings of all the variables in that production.  
;;; It needs to be updated to handle tutor mode, but again, for now I just
;;; need to get it working.

(defun stepper-instan-binding (prod)
  (when prod
    (if (equal *which-stepper-list* 'prod)
        (let* ((instan (find prod *conflict-set* 
                             :key #'(lambda (x) 
                                      (production-name 
                                       (instantiation-production x)))))
               (selected (equal instan (car *conflict-set*)))
               (bindings (make-ie-instantiation instan))
               (width (apply #'max (mapcar #'(lambda (x) 
                                               (length (string (first x))))
                                     bindings))))
          
          (if (and selected *tutor-mode* (eq :instan *which-hook-fn*))
              (progn
                (when (null *tutor-bindings*)
                  (setf *tutor-bindings* (copy-tree bindings))
                  (setf *tutor-responses* (mapcar #'(lambda (x) (cons (car x) nil)) *tutor-bindings*)))
                (dolist (x *tutor-responses*)
                  (format t "~vs: ~a~%" (1+ width) (car x) (if (null (cdr x)) "{binding}" (cdr x)))))
            (dolist (x bindings)
              (format t "~vs: ~s~%" (1+ width) (car x) (cdr x)))))
      (progn
        (format t "Sources: ~S~%" (cadar *retrieval-record*))
        (format t "Retrieval Request:~%")
        (let ((clause (copy-tree (caar *retrieval-record*))))
          (format t "~S~%" (first clause))
          (dolist (slot-value (rest clause))
            (if (null (rest slot-value))
                (format t "   ~S~%" (first slot-value))
              (format t " ~A ~S ~S~%"
                (if (test-modifier (first slot-value)) 
                    (pop slot-value) 
                  " ")
                (first slot-value)
                (second slot-value)))))))))
           
        

;;; stepper-instan-production
;;; This is the function that handles the updates of the instantiation
;;; produciton box in the stepper.  The parameter passed in is the name of
;;; a production that's chosen in the list (that's the only way this is called)
;;; It prints the instantiated produciton.

(defun stepper-instan-production (prod)
  (when prod
    (if (equal *which-stepper-list* 'prod)
        (let ((instan (find prod *conflict-set* 
                            :key #'(lambda (x) 
                                     (production-name 
                                      (instantiation-production x))))))
          
          (my-pprint-instantiation instan))
      (wm-fct (list prod)))))


;;; These next 3 functions are from my old environment.  
;;; They are copied from the ACT-R functions of the same name without the 
;;; "my-" on the front.
;;; The main reason for using my own versions instead of the originals is
;;; that the originals used signal-output that has a forced case and "depth"
;;; associated with it that I don't want.

(defmacro my-print-side (side)
  "Prints one side of a production, 
   i.e. a list of commands and wme retrievals."
  `(dolist (clause ,side)
     (cond ((retrievalp (first clause))
            (format trace "   ~S~%" (first clause) )
            (dolist (slot-value (rest clause))
              (if (null (rest slot-value))
                (format trace "      ~S~%" (first slot-value))
                (format trace "    ~A ~S ~S~%"
                  (if (test-modifier (first slot-value)) 
                      (pop slot-value) 
                    " ")
                  (first slot-value)
                  (second slot-value)))))
           (t
            (format trace "  ~{ ~S~}~%" clause)))))

(defun my-pprint-instantiation (&optional (instantiation *instantiation*)
                                       (trace *command-trace*))
  "Pretty-prints an instantiation."
  (let ((selected (equal instantiation (car *conflict-set*)))
        (production (instantiation-production instantiation)))
    (let ((bindings nil)
          (text nil))
      (if (and selected *tutor-mode* (eq :instan *which-hook-fn*))
          (let ((bound (remove-if #'null *tutor-responses* :key #'cdr)))
            (setf bindings (append bound (mapcar #'(lambda (x) (cons (var-var> (car x)) (var-var> (cdr x)))) bound))))
      
      (dolist (binding (production-bindings production))
        (when (integerp (variable-index binding))
          (let ((value (instantiation-variable instantiation 
                                               (variable-index binding))))
            
              (when value
                (push (cons (variable-name binding) value) bindings)
                (push (cons (var-var> (variable-name binding))
                            (var-var> value))
                      bindings))))))
      (setf text (sublis bindings (production-text production)))
      (my-pprint-production production trace text))))

(defun my-pprint-production (production 
                             &optional (trace *command-trace*)
                             (text (production-text production))
                             (documentation 
                              (production-documentation production)))
  "Pretty prints production.  Pretty basic."
  (format trace "(p ~S~%" production )
  (when documentation
    (format trace "   ~S~%" documentation ))
  (my-print-side (car text))
  (format trace "==>~%" )
  (my-print-side (cdr text))
  (format trace ")"))

;;; make-ie-instantiation
;;; This function takes one parameter which should be an instantiation.
;;; It returns a list of conses where each cons is a variable in the production
;;; and its current binding for all of the variables in the produciton.

(defun make-ie-instantiation (inst)
  (mapcar  #'(lambda (binding)
               (cons (variable-name binding)
                     (if (integerp (variable-index binding))
                         (let ((temp (instantiation-variable 
                                      inst (variable-index binding))))
                           (if (wme-p temp)
                               (wme-name temp)
                             (if (null temp)
                                 :unbound
                               temp)))
                       :return-value-variable)))
    
    (remove-if-not #'(lambda (x) 
                       (equal (aref (string (variable-name x)) 0) #\=))
                   (production-bindings (instantiation-production inst)))))


;;; currently-handled-event
;;; This function is used to check whether the "current" event in the queue
;;; should stop the system for inspection based on the settings of the 
;;; checkboxes in the stepper dialog.  The event will be either a queue-
;;; entry or an instantiation, and if it's a queue entry there are
;;; some things that need to be checked because there's an new hook that
;;; makes things "easier".

(defvar *foo*)

(defun currently-handled-event (event)
  (if event
      (let ((event-type (if (subtypep (type-of event) 'queue-entry)
                            (destination event) 
                          :production)))
        (cond 
         ;; always handle a post event
         ((eq :post *which-hook-fn*) t)
         ;; if it's fast-forwarding check it
         ((numberp *stepper-skip-until-time*)
          (if (>= (mp-time *mp*) *stepper-skip-until-time*)
              (progn 
                (setf *stepper-skip-until-time* nil)
                t)
            nil))
         ;; handle the skip to production situation
         (*stepper-skip-until-time* ;; it's a symbol for a production name
          (if (and (eq :instan *which-hook-fn*)
                   (car event)
                   (instantiation-production (car event))
                   (equal *stepper-skip-until-time* 
                          (production-name (instantiation-production (car event)))))
              (progn
                (setf *stepper-skip-until-time* nil)
                t)
              nil))
         ;; cognition needs some special care
         ((eq event-type :cognition)
          (if (eq :pre *which-hook-fn*)
              nil ;; don't worry about the pre cuz there'll be a :step
            (let ((sub-command (car (params event))))
              (if (equal sub-command 'complete-retrieval)
                  (gethash :retrieval *stepper-events*)
                (if (equal sub-command 'select-production)
                    nil ;; skip it cuz the instan will catch it
                  (gethash :production *stepper-events*))))))
         ((and (eq :pre *which-hook-fn*)
               (not (eq :external event-type)))
          nil)
         ((not (eq :pre *which-hook-fn*))
          (gethash event-type *stepper-events*))
         (t ;; it's got to be :pre and :external
          (if (equal (car (params event)) 'stepper-continuing)
              t
            (gethash event-type *stepper-events*)))))
    nil))



;;; run-sched-queue
;;; This after method on run-sched-queue is used to clean up after my
;;; "stopping" system...  To stop the model in the stepper I push a bogus
;;; event onto the head of the schedule queue in the post event hook so 
;;; I need to clean that up here.

(defmethod run-sched-queue :after ((mstr-proc master-process) stop-time)
  (declare (ignore stop-time))
  (when *stopped-by-stepper*
    (pop (sched-q mstr-proc))))

;;; run-master-process
;;; Didn't want to do things this way, but seems easiest
;;; Add a stopping condition to the cond block that handles
;;; terminating if *stopped-by-stepper* is set

(defmethod run-master-process ((mstr-proc master-process) duration)
  "Run a Master Process duration seconds."
  (setf (sim-start-real mstr-proc) (get-internal-real-time))
  (setf (start-time mstr-proc) (mp-time mstr-proc))
  (let ((current-time (mp-time mstr-proc))
        (finish-time (ms-round (+ duration (mp-time mstr-proc))))
        (cog-mod (cog-m mstr-proc))
        )
    (while (< (mp-time mstr-proc) finish-time)
      (let ((ps-status (ps-state cog-mod)))
        (cond     
         ;; If the stop button was pressed just quit.
         (*stopped-by-stepper*
          (setf *stopped-by-stepper* nil)
          (pm-output nil 
                     "* System stopped by the stepper.")
          (return-from run-master-process
            (values (elapsed-sim-time mstr-proc) current-time)))
         ;; pending execution--a production has matched and the RHS has been
         ;; scheduled, so there's guranteed to be a next event
         ((eq ps-status :PENDING-EXECUTION) 
          (setf current-time (min finish-time (next-stop-time mstr-proc)))
          (do-update mstr-proc current-time :real-wait t))
         
         ;; If there's no goal in focus, ACT won't ever do anything.  Bail.
         ((eq ps-status :NO-GOAL)
          (pm-output nil 
                     "* Production system halted because GOAL is cleared.")
          (maybe-dequeue-events mstr-proc finish-time)
          (setf (ps-state cog-mod) :IDLE)
          (return-from run-master-process
            (values (elapsed-sim-time mstr-proc) current-time)))
         ;; If no production matched and there are pending events, then
         ;; just keep waiting until the next event.
         ((and (eq ps-status :FAILED-SELECTION)
               (sched-q mstr-proc))
          (setf (ps-state cog-mod) :IDLE)
          (setf current-time (min finish-time
                                  (stop-time mstr-proc finish-time)))
          (do-update mstr-proc current-time :real-wait t))
         ;; If no production matched and there are no pending events, either
         ;; nothing will ever happen and we're done, or there are silent 
         ;; events and we should keep trying.
         ((eq ps-status :FAILED-SELECTION)
          (if (no-silent-events mstr-proc)
            ;; no silent events: nothing will happen, bail out
            (progn 
              (setf (ps-state cog-mod) :IDLE)
              (pm-output nil 
                         "* Nothing to run:  No productions, no events.")
              (return-from run-master-process
                (values (elapsed-sim-time mstr-proc) current-time))
              )
            ;; silent events--try another cognition cycle.
            (progn
              (queue-command :time (min-run-time mstr-proc) :where :COGNITION
                             :command 'select-production
                             :params (list 1)
                             :priority -1)
              (do-update mstr-proc current-time :real-wait t))
            ))        
         ;; This state means a production cycle has completed, so it's time
         ;; to schedule another cycle.
         ((eq ps-status :IDLE) 
          ;(schedule-ps current-time finish-time)
          (queue-command :time 0 :where :COGNITION
                         :command 'select-production
                         :params (list 1)
                         :priority -1)
          (do-update mstr-proc current-time :real-wait t))
         ;; fail!
         (t (error 
             "Unhandled state--email byrne@acm.org or db30@andrew.cmu.edu!")))
        ))
    (values (elapsed-sim-time mstr-proc) current-time)))


(defun tutored-step ()
  (and *tutor-mode* (eq :instan *which-hook-fn*))
  )

(defun tutor-completed (word)
  (if (cdr (assoc word *tutor-responses*))
      1 0))

(defun tutor-check (word binding)
  (if (equal binding (cdr (assoc word *tutor-bindings*)))
      (progn
        (setf (cdr (assoc word *tutor-responses*)) binding)
        1)
    0))

(defun tutor-answer (word)
  (cdr (assoc word *tutor-bindings*)))
#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#